from sympy import *
import mpmath

R12 = Rational(1,2)
R32 = Rational(3,2)
R52 = Rational(5,2)
R14 = Rational(1,4)
R34 = Rational(3,4)

_sum = __builtins__.sum
fac = factorial

def ssqrt(z):
    """
    Cheating square root, to evaluate formulas of type 
    F(z^2) -> f(z) (TODO: this is fine sometimes, but not always)
    """
    p = Integer(1)
    if z.is_Mul:
        return Mul(*(ssqrt(r) for r in z.args))
    try:
        d = z.as_powers_dict()
        for base, expt in d.items():
            p *= base ** (expt / 2)
    except (AttributeError, TypeError):
        p = sqrt(z)
    return p

def _0f1(b,z):
    """
    Evaluation of 0F1(b,z), assuming b a half-integer
    """
    F12 = cosh(2*ssqrt(z))
    F32 = sinh(2*ssqrt(z))/(2*ssqrt(z))
    if b == R12: return F12
    if b == R32: return F32
    if b > R32:
        return (b-2)*(b-1)/z * (_0f1(b-2,z) - _0f1(b-1,z))
    if b < R12:
        return _0f1(b+1,z) + z/(b*(b+1)) * _0f1(b+2,z)
    raise ValueError

def _2f1(a,b,c,z):
    """
    Evaluation of 2F1(a,b,c,z), assuming a,b,c positive
    integers or half-integers
    """
    if b == c:
        return (1-z)**(-a)
    if a == c:
        return (1-z)**(-b)
    if a == 0 or b == 0:
        return Integer(1)
    if a > b:
        a, b = b, a
    if b >= 2:
        F1 = _2f1(a,b-1,c,z)
        F2 = _2f1(a,b-2,c,z)
        q = (b-1)*(z-1)
        return ((c-2*b+2+(b-a-1)*z)*F1 + (b-c-1)*F2) / q
    if c > 2:
        # how to handle this case?
        if a-c+1 == 0 or b-c+1 == 0:
            raise NotImplementedError
        F1 = _2f1(a,b,c-1,z)
        F2 = _2f1(a,b,c-2,z)
        r1 = (c-1)*(2-c-(a+b-2*c+3)*z)
        r2 = (c-1)*(c-2)*(1-z)
        q = (a-c+1)*(b-c+1)*z
        return (r1*F1 + r2*F2) / q
    # TODO: elliptic integrals
    if (a,b,c) == (R12,R12,1): return 2*EllipticK(z)/pi
    if (a,b,c) == (R12,R12,R32): return asin(ssqrt(z))/ssqrt(z)
    if (a,b,c) == (R12,R12,2): return 4*(EllipticE(z)+(z-1)*EllipticK(z))/(pi*z)
    if (a,b,c) == (R12,1,R32): return atanh(ssqrt(z))/ssqrt(z)
    if (a,b,c) == (R12,1,2): return (2-2*sqrt(1-z))/z
    if (a,b,c) == (R12,R32,1): return -2*EllipticE(z)/(pi*(z-1))
    if (a,b,c) == (R12,R32,2): return -r*(EllipticE(z)-EllipticK(z))/(pi*z)
    if (a,b,c) == (1,1,R12): return (-ssqrt(1-z)-ssqrt(z)*asin(ssqrt(z)))/(ssqrt(1-z)*(z-1))
    if (a,b,c) == (1,1,R32): return asin(ssqrt(z))/(ssqrt(1-z)*ssqrt(z))
    if (a,b,c) == (1,1,2): return -log(1-z)/z
    if (a,b,c) == (1,R32,R12): return (1+z)/(1-z)**2
    if (a,b,c) == (1,R32,2): return 2*(1/ssqrt(1-z)-1)/z
    if (a,b,c) == (1,2,R12): return (2+z+3*ssqrt(z)*asin(ssqrt(z))/ssqrt(1-z))/(2*(z-1)**2)
    if (a,b,c) == (1,2,R32): return (-(ssqrt(1-z)*ssqrt(z))-asin(ssqrt(z)))/(2*ssqrt(1-z)*(z-1)*ssqrt(z))
    if (a,b,c) == (R32,2,R12): return (1+3*z)/(1-z)**3
    if (a,b,c) == (R32,2,1): return (2+z)/(2*(ssqrt(1-z)*(1-z)**2))
    if (a,b,c) == (2,2,R12): return (ssqrt(1-z)*(4+11*z)+3*ssqrt(z)*(3+2*z)*asin(ssqrt(z))) / (4*(ssqrt(1-z)*(1-z)**3))
    if (a,b,c) == (2,2,1): return (1+z)/(1-z)**3
    if (a,b,c) == (2,2,R32): return (3+(1+2*z)*asin(ssqrt(z))/ssqrt(-z*(z-1)))/(4*(z-1)**2)
    raise ValueError

def polygammax(n,z):
    """
    Polygamma function with some evaluations currently not done
    automatically by SymPy's polygamma
    """
    if n == 0:
        if (not z.is_Integer) and z.is_Rational:
            p, q = z.as_numer_denom()
            s = -EulerGamma - log(q) - pi/2/tan(pi*p/q)
            for k in xrange(1,int(q)):
                s += cos(2*pi*k*p/q)*log(2-2*cos(2*pi*k/q))/2
            return s
    if n % 2 == 0 and z.is_Integer and z > 0:
        if z != 1:
            return polygammax(n,z-1) + (-1)**n * fac(n) * (z-1)**-(n+1)
        return zeta(n+1)*fac(n)*(-1)**(n+1)
    if n == 1 and z == R12: return pi**2/2
    if n == 1 and z == R14: return 8*Catalan + pi**2
    if n == 1 and z == R34: return -8*Catalan + pi**2
    return polygamma(n,z)

def ratsum(aa, bb):
    """
    Sum (k+a_1)...(k+a_p) / ((k+b_1)...(k+b_q)) for k=0...inf.
    TODO: implement general case, e.g. using partial fraction decomposition
    """
    p = len(aa)
    q = len(bb)
    if p == 0 and len(set(bb)) == 1:
        return polygammax(q-1, bb[0]) * (-1)**q / fac(q-1)
    if p == 0 and q == 2:
        b1, b2 = bb
        return (polygammax(0,b1)-polygammax(0,b2))/(b1-b2)
    if p == 1 and len(set(bb)) == 1:
        a = aa[0]
        b = bb[0]
        return (-1)**q * ((a-b)*polygammax(q-1,b) - (q-1)*polygammax(q-2,b)) / fac(q-1)
    raise NotImplementedError

def altratsum(aa, bb):
    p = len(aa)
    q = len(bb)
    if p == 0 and len(set(bb)) == 1:
        b = bb[0]
        return (-1)**(q+1) * (polygammax(q-1,(1+b)/2) - polygammax(q-1,b/2))/(fac(q-1)*2**q)
    raise NotImplementedError

def try_ratsum(aa, bb, z):
    """
    Convert and evaluate hypergeometric series into rational
    series and evaluate, if possible.
    TODO: handle other powers of unity than 1
    TODO: also handle rational-geometric series using
    polylogarithms, Lerch transcendents...
    """
    if z not in (1,-1) or len(aa) != len(bb):
        raise NotImplementedError
    coeff = Integer(1)
    pairs = []
    RA = []
    RB = []
    for a in aa:
        i = 0
        while i < len(bb):
            b = bb[i]
            n = b-a
            if n.is_Integer:
                pairs.append((a,b,n))
                del bb[i]
                break
            else:
                i += 1
    if len(pairs) != len(aa):
        raise NotImplementedError
    for (a,b,n) in pairs:
        if n < 0:
            coeff /= rf(b,-n)
            for k in xrange(int(-n)):
                RA.append(b+k)
        else:
            coeff *= rf(a,n)
            for k in xrange(int(n)):
                RB.append(a+k)
    if z == 1:
        return coeff * ratsum(RA, RB)
    if z == -1:
        return coeff * altratsum(RA, RB)
    raise NotImplementedError

def contiguous(h):
    """
    Try to reduce hypergeometric function to a sum of contiguous
    functions so as to reduce the degree.
    """
    if not isinstance(h, Hyper):
        return h
    p,q,aa,bb,z = h.p,h.q,h.a,h.b,h.z
    aa = list(aa)
    bb = list(bb)
    if q < 1 or p < 2:
        return h
    for i,w in enumerate(aa):
        for j,c in enumerate(bb):
            # increase denominator parameter until equal
            # to a numerator parameter
            n = w-c
            if n.is_Integer and n > 0:
                a = aa[(i+1)%p]
                aa1, bb1 = aa[:], bb[:]
                aa2, bb2 = aa[:], bb[:]
                aa1[(i+1)%p] += 1
                bb1[j] += 1
                bb2[j] += 1
                F1 = Hyper(aa1, bb1, z)
                F2 = Hyper(aa2, bb2, z)
                return (a*F1 - (a-c)*F2)/c
    return h

class Hyper(Basic):
    def __new__(cls, a, b, z):
        self = object.__new__(cls)
        self_a = sorted(map(sympify, a))
        self_b = sorted(map(sympify, b))
        self_z = z = sympify(z)
        # degree reduction
        p = len(self_a)
        q = len(self_b)
        i = 0
        while i < q and self_a:
            b = self_b[i]
            if b in self_a:
                self_a.remove(b)
                self_b.remove(b)
                p -= 1
                q -= 1
            else:
                i += 1
        # Trivial cases
        if z == 0:
            return Integer(1)
        if p == q == 0:
            return exp(z)
        if p == 1 and q == 0:
            return (1-z)**(-self_a[0])
        if p == 0 and q == 1:
            b = self_b[0]
            if (b-R12).is_Integer:
                return _0f1(b,z)
        # Confluent hypergeometric function
        if p == 1 and q == 1:
            a, b = self_a[0], self_b[0]
            if a == R12 and b == R32:
                C, factors = z.as_coeff_terms()
                if C < 0:
                    t = ssqrt(-C)*ssqrt(Mul(*factors))
                    return (sqrt(pi)/2) * erf(t) / t
            if a == 1 and b == 2:
                return (exp(z)-1)/z
            n, m = a, b
            if n.is_Integer and m.is_Integer and m > 0 and n > 0:
                if m <= n:
                    return exp(z)*_sum(rf(m-n,k)*(-z)**k/fac(k)/rf(m,k) for k in xrange(n-m+1))
                else:
                    T = _sum(rf(n-m+1,k)*z**k/(fac(k)*rf(2-m,k)) for k in xrange(m-n))
                    U = _sum(rf(1-n,k)*(-z)**k/(fac(k)*rf(2-m,k)) for k in xrange(n))
                    return fac(m-2)*rf(1-m,n)*z**(1-m)/fac(n-1)*(T - exp(z)*U)
        # Gauss hypergeometric function
        if p == 2 and q == 1:
            a, b = self_a; c, = self_b
            if z == 1:
                return gamma(c)*gamma(c-a-b)/gamma(c-a)/gamma(c-b)
            if (a*2).is_Integer and (b*2).is_Integer and (c*2).is_Integer and a > 0 and b > 0 and c > 0:
                try:
                    return _2f1(a,b,c,z)
                except NotImplementedError:
                    pass
        try:
            return try_ratsum(self_a, self_b + [Integer(1)], z)
        except NotImplementedError:
            pass
        args = (p, q, tuple(self_a), tuple(self_b), z)
        v = Basic.__new__(cls, *args)
        return contiguous(v)
    @property
    def p(self):
        return self.args[0]
    @property
    def q(self):
        return self.args[1]
    @property
    def a(self):
        return self.args[2]
    @property
    def b(self):
        return self.args[3]
    @property
    def z(self):
        return self.args[4]
    def _sympystr_(self, *args):
        return "%iF%i(%s, %s, %s)" % (self.p, self.q, list(self.a), list(self.b), self.z)
    def _eval_evalf(self, prec):
        # TODO: rational parameters, scaling of z...
        orig = mpmath.mp.prec
        try:
            mpmath.mp.prec = prec
            def tompmath(z):
                try:
                    return mpmath.mpf(z)
                except:
                    return mpmath.mpf(re(z)) + mpmath.mpf(im(z))*1j
            aa = [tompmath(a._eval_evalf(prec)) for a in self.a]
            bb = [tompmath(b._eval_evalf(prec)) for b in self.b]
            z = tompmath(self.z._eval_evalf(prec))
            v = mpmath.hyper(aa,bb,z)
            y = sympify(v.real) + sympify(v.imag)*I
        finally:
            mpmath.mp.prec = orig
        return y
    def _eval_subs(self, old, new):
        aa = self.a
        bb = self.b
        z = self.z
        aa = [a._eval_subs(old,new) for a in aa]
        bb = [b._eval_subs(old,new) for b in bb]
        z = z._eval_subs(old,new)
        return Hyper(aa,bb,z)
    def _latex_(self, *args):
        p, q = self.p, self.q
        aa = ",".join([latex(a) for a in self.a])
        bb = ",".join([latex(b) for b in self.b])
        z = latex(self.z)
        s = r"\,_{%s}F_{%s}\left(%s; %s; %s\right)"
        s = (s % (p, q, aa, bb, z)).replace("$", "")
        return s

def hypsum(expr, var):
    var = sympify(var)
    expr = sympify(expr)
    expr = together(expr)
    constant_factor, factors = expr.as_coeff_terms(var)
    numer = [Integer(1)]
    denom = []
    power = Integer(1)
    a, b = Wild('a', exclude=[var]), Wild('b', exclude=[var])
    linear = a + b*var
    factors = list(factors)
    for factor in factors:
        for subfactor, exponent in factor.as_powers_dict().items():
            if isinstance(subfactor, Binomial):
                subfactor = Binomial(*subfactor.args)
            if isinstance(subfactor, Factorial):
                subfactor = Gamma(subfactor.args[0]+1)
            if isinstance(subfactor, exp):
                # exp(a+bk) = exp(a) exp(bk) = exp(a) exp(b)^k
                w = subfactor.args[0]
                m = expand(w).match(linear)
                if m is None:
                    raise ValueError("not hypergeometric")
                constant_factor *= exp(m[a])
                subfactor = exp(m[b])
                exponent = var
            if isinstance(subfactor, (cos,sin)):
                terms = expand(expr.rewrite(exp)).as_coeff_factors()[1]
                return _sum(hypsum(term,var) for term in terms)
            if exponent.is_Integer:
                if isinstance(subfactor, RisingFactorial):
                    x, n = subfactor.args
                    xm = x.match(linear)
                    nm = n.match(linear)
                    if xm is None or nm is None:
                        raise ValueError("not hypergeometric")
                    # (A+B*k)_(C+D*k)
                    A = xm[a]; B = xm[b]
                    C = nm[a]; D = nm[b]
                    if B == 0 and C == 0 and D == 1:
                        if exponent > 0:
                            for n in range(int(exponent)): numer.append(A)
                        else:
                            for n in range(-int(exponent)): denom.append(A)
                    else:
                        raise NotImplementedError
                elif isinstance(subfactor, gamma):
                    x = subfactor.args[0]
                    xm = x.match(linear)
                    if xm is None:
                        raise ValueError("not hypergeometric")
                    # gamma(A + B*k)
                    A = xm[a]; B = xm[b]
                    if B == 1:
                        constant_factor *= gamma(A) ** exponent
                        if exponent > 0:
                            for n in range(int(exponent)): numer.append(A)
                        else:
                            for n in range(-int(exponent)): denom.append(A)
                    elif B.is_Integer and B > 0:
                        c = B**(B*var+A-Rational(1,2))*(2*pi)**((1-B)/2)
                        expr2 = expr / factor
                        expr2 *= c**exponent
                        for k in range(int(B)):
                            expr2 *= gamma(var+(A+k)/B) ** exponent
                        return hypsum(expr2, var)
                    else:
                        raise NotImplementedError
                else:
                    # polynomial
                    P = Poly(subfactor, var).as_monic().as_basic()
                    C = quo(P, subfactor, var)
                    assert not C.has(var)
                    constant_factor *= C ** -exponent
                    subfactor = P
                    for r, rexp in roots(subfactor, var).items():
                        A = -r
                        if A == 0:
                            return hypsum(expr.subs(var, var+1), var)
                        else:
                            e = rexp * exponent
                            constant_factor *= A ** e
                            if e > 0:
                                for n in range(int(e)): numer.append(A+1)
                                for n in range(int(e)): denom.append(A)
                            else:
                                for n in range(-int(e)): denom.append(A+1)
                                for n in range(-int(e)): numer.append(A)
            elif exponent.has(var) and not subfactor.has(var):
                m = exponent.match(linear)
                if m is None:
                    raise ValueError("not hypergeometric")
                constant_factor *= subfactor ** m[a]
                power *= subfactor ** m[b]
            else:
                raise ValueError("not hypergeometric")
    return constant_factor * Hyper(numer, denom, power)

k = Symbol('k', positive=True)
d = Symbol('d')
z = Symbol('z')

if __name__ == "__main__":
    print hypsum(z**k / fac(k), k)
